"""
xenops.data.converter
~~~~~~~~~~~~~~~~~~~~~

:copyright: 2017 by Maikel Martens
:license: GPLv3
"""


class BaseConverter:
    """Base converter"""

    def __init__(self, attribute, service_attribute):
        """
        Init BaseConverter

        :param str attribute:
        :param str service_attribute:
        """
        self.attribute = attribute
        self.service_attribute = service_attribute

    def import_attribute(self, data):
        """
        Convert raw service data to DataType data

        :param dict data:
        :return:
        """
        keys = self.service_attribute.split('.')
        return self.get_import_value(keys, data)

    def get_import_value(self, keys, data):
        """
        Recursive function for getting data from service data dict

        :param str keys:
        :param dict data:
        :return:
        """
        key = keys[0]
        keys = keys[1:]

        # TODO: Check if key is array value like: items[] or items[1]

        if not keys:
            if key not in data:
                raise KeyError()
            return data.get(key)
        if key in data:
            return self.get_import_value(keys, data.get(key))

    def export_attribute(self, data_object):
        """
        Convert DataType data to service data

        :param xenops.data.DataType data_object:
        :return:
        """
        return data_object.get(self.attribute)

    def __str__(self):
        """
        Representation of converter

        :return str:
        """
        return '{}: {{attribute: {},  service_attribute: {}}}'.format(
            self.__class__.__name__,
            self.attribute,
            self.service_attribute
        )


class Attribute(BaseConverter):
    """Default attribute converter"""

    pass



"""
xenops.data
~~~~~~~~~~~

:copyright: 2017 by Maikel Martens
:license: GPLv3
"""
import logging

logger = logging.getLogger(__name__)


class DataTypeFactory:
    """Datatype factory for registering and getting data types"""

    MODE_MERGE = 'merge'
    MODE_REPLACE = 'replace'

    _datatypes = {}

    @classmethod
    def register(cls, code, attributes, mode=MODE_MERGE):
        # TODO: validate attributes
        """
        Register a date type.

        :param str code:
        :param list attributes:
        :param str mode:
        """
        logger.info("TODO: validate attributes")

        if code in cls._datatypes:
            datatype = cls._datatypes[code]

            if mode == cls.MODE_MERGE:
                for attribute_code, attribute_config in attributes.items():
                    current_config = datatype.attributes.get(attribute_code, {})
                    for key, value in attribute_config.items():
                        current_config[key] = value
                    datatype.attributes[attribute_code] = current_config
            else:
                datatype.attributes = attributes
        else:
            datatype = DataType(code, attributes)

        cls._datatypes[code] = datatype

    @classmethod
    def get(cls, code):
        """
        Get DataType object based on data type code

        :param str code:
        :return xenops.data.DataType:
        """
        return cls._datatypes.get(code)


class DataType:
    """DataType class"""

    def __init__(self, code, attributes, verbose_name=''):
        """
        Init DataType

        :param str code:
        :param dict attributes:
        :param str verbose_name:
        """
        self.code = code
        self.attributes = attributes
        self.verbose_name = verbose_name if verbose_name else code.replace('_', '').title()

    def is_valid_attribute(self, code):
        """
        Check if given code is a valid attribute code for data type

        :param str code:
        :return bool:
        """
        return code in self.attributes

    def is_valid_attribute_value(self, code, value):
        """
        Validate if value is an valid value for given attribute code

        :param str code:
        :param value:
        :return bool:
        """
        # TODO: Add validation + spec out the attribute config
        return True


class DataMapObject:
    """DataMapObject for converting raw service data to datatype data"""

    def __init__(self, datatype, mapping, enhancers, data):
        """
        Init DataMapObject

        :param xenops.data.DataType datatype:
        :param dict mapping:
        :param list enhancers:
        :param dict data:
        """
        self.datatype = datatype
        self.mapping = mapping if mapping else {}
        self.enhancers = enhancers if enhancers else []
        self.data = data if data else {}
        self.cached_mapping_data = {}

        for enhancer in self.enhancers:
            enhancer.source_object = self

    def get(self, key, default=None, raise_keyerror=False):
        """
        Get value from DataMapObject with datatype attribute code

        If there is an mapping for given key then will try to get value with that mapping.
        If there is no mapping or value it will check if one of the enhancers has the value.

        :param str key:
        :param default:
        :param bool raise_keyerror:
        :return:
        """
        if key in self.cached_mapping_data:
            return self.cached_mapping_data.get(key)

        if key in self.mapping:
            # @TODO Validate value is_valid_attribute_value
            try:
                value = self.mapping.get(key).import_attribute(self.data)
                self.cached_mapping_data[key] = value
                return value
            except KeyError:
                pass

        # try get value from enhancers
        for enhancer in self.enhancers:
            if enhancer.provide_attribute(key):
                value = enhancer.get(key, default, raise_keyerror)
                self.cached_mapping_data[key] = value
                return value

        if raise_keyerror:
            raise KeyError('Attribute ({}) does not exists'.format(key))

        return default

    def export_to(self, mapping):
        """
        Convert datatype date to given mapping

        :param mapping:
        :return dict:
        """
        # TODO
        return {}


class Enhancer:
    """Enhancer class"""

    def __init__(self, connector, mapping, attributes):
        """
        Init Enhancer class

        :param xenops.connector.Connector connector:
        :param dict mapping:
        :param list attributes:
        """
        self.connector = connector
        self.mapping = mapping
        self.attributes = attributes
        self.source_object = None
        self.data = None
        self.cached_mapping_data = {}

    def provide_attribute(self, attribute_code):
        """
        Check if given attribute code is provided by enhancer

        :param str attribute_code:
        :return:
        """
        return attribute_code in self.attributes

    def get(self, attribute_code, default=None, raise_keyerror=False):
        """
        Get value for given attribute code

        :param str attribute_code:
        :param default:
        :param bool raise_keyerror:
        :return:
        """
        if not self.data:
            self._load_data()

        return self.data.get(attribute_code, default, raise_keyerror)

    def _load_data(self):
        """Load DataMapObject"""
        self.data = self.connector.get(self.source_object.datatype.code, self.source_object.get('id'))



"""
trionyx.conf
~~~~~~~~~~~~

:copyright: 2017 by Maikel Martens
:license: GPLv3
"""
import os
import importlib
import logging

logger = logging.getLogger(__name__)

ENVIRONMENT_VARIABLE = "XENOPS_SETTINGS"


class Settings:
    """Settings wrapper class"""

    def __init__(self):
        """Init setting and load settings from project"""
        self.IS_PROJECT = False
        self._settings = {}

        project_settings = os.environ.get(ENVIRONMENT_VARIABLE)
        if project_settings:
            self._settings = importlib.import_module(project_settings)
            self.IS_PROJECT = True

    def __getattr__(self, name):
        """Get setting value"""
        if hasattr(self._settings, name):
            return getattr(self._settings, name, None)

    def get(self, key, default=None):
        """
        Get settings with default

        :param str key:
        :param default:
        :return:
        """
        if hasattr(self._settings, key):
            return getattr(self._settings, key, None)
        return default


settings = Settings()



"""
xenops.app
~~~~~~~~~~

:copyright: 2017 by Maikel Martens
:license: GPLv3
"""
import logging
from pkg_resources import iter_entry_points

from xenops.conf import settings
from xenops.service import ServiceFactory
from xenops.data import DataTypeFactory
from xenops.data.types import default_types
from xenops.connector import Connector, InvalidConnectorConfig

logger = logging.getLogger(__name__)


class Application:
    """Application holds all project data"""

    def __init__(self):
        """Load all services and project settings"""
        self.connectors = {}

        # Load all default product types
        for code, attributes in default_types.items():
            DataTypeFactory.register(code, attributes)

        self.load_services()

        if settings.IS_PROJECT:
            logger.debug("Loading project settings")
            self.load_project_types()
            self.load_project_connectors()

    def load_services(self):
        """Load all services registered with pkg_resources"""
        for entry_point in iter_entry_points(group='xenops.services', name=None):
            try:
                config = entry_point.load()
            except ImportError:
                logger.error('Could not load service: ({}), check your setup.py entry_points.'.format(entry_point))
                continue

            if type(config) != dict:
                logger.error('Service config ({}) is not an dict and will not be registered!'.format(entry_point))

            logger.debug("Registering ({}) service".format(entry_point))
            if not ServiceFactory.register(config):
                logger.error('Service ({}) could not be registered!'.format(entry_point))

    def load_project_types(self):
        """Load project types"""
        for code, config in settings.get('TYPES', {}).items():
            # TODO add validation
            DataTypeFactory.register(code, config['attributes'], config.get('type', 'merge'))

    def load_project_connectors(self):
        """Load project connectors"""
        for code, config in settings.get('CONNECTORS', {}).items():
            config['code'] = code
            try:
                self.connectors[code] = Connector.create_connector(self, config)
            except InvalidConnectorConfig as e:
                logger.error('Invalid connector ({}) config: {}'.format(code, e))

    def trigger(self, type_code=None, connector_code=None):
        """
        Trigger a connector data type import

        :param str type_code:
        :param str connector_code:
        """
        if not DataTypeFactory.get(type_code):
            raise Exception('Invalid type')

        if connector_code not in self.connectors:
            raise Exception('Connector does not exists')

        connector = self.connectors.get(connector_code)
        connector.execute_trigger(type_code)



"""
xenops.data.types.product
~~~~~~~~~~~~~~~~~~~~~~~~~

:copyright: 2017 by Maikel Martens
:license: GPLv3
"""
product = {
    'sku': {
        'type': str,
        'verbose_name': 'Sku',
        'allowed_value': r'[\w\d]+',
    },
    'price': {
        'type': float,
    }
}



"""
xenops.connector
~~~~~~~~~~~~~~~~

:copyright: 2017 by Maikel Martens
:license: GPLv3
"""
import logging
import datetime

from xenops.data.converter import BaseConverter
from xenops.service import ServiceFactory, TriggerRequest, GetRequest, ProcessRequest
from xenops.data import DataMapObject, Enhancer

logger = logging.getLogger(__name__)


class InvalidConnectorConfig(Exception):
    """Invalid Connector config Exception"""

    pass


class InvalidCode(Exception):
    """Invalid code Exception"""

    pass


class Connector:
    """Connector"""

    def __init__(self, app, code, service, verbose_name, service_config=None, mapping=None, triggers=None,
                 enhancers=None, processes=None):
        """
        Init Connector

        :param xenops.app.Application app:
        :param str code:
        :param xenops.service.Service service:
        :param str verbose_name:
        :param dict service_config:
        :param dict mapping:
        :param dict triggers:
        :param list enhancers:
        :param list processes:
        """
        self.app = app
        self.code = code
        self.service = service
        self.verbose_name = verbose_name if verbose_name else code
        self.service_config = service_config if service_config else {}
        self.mapping = mapping if mapping else {}
        self.triggers = triggers if triggers else {}
        self.enhancers = enhancers if enhancers else []
        self.processes = processes if processes else []

    @classmethod
    def create_connector(cls, app, config):
        """
        Create connector based on config

        :param xenops.app.Application app:
        :param dict config:
        :return Connector:
        """
        return cls(
            app,
            **ConnectorConfig().parse(config)
        )

    def execute_trigger(self, trigger_code):
        """
        Run trigger process based on trigger code

        :param str trigger_code:
        """
        trigger = self.triggers.get(trigger_code)
        if not trigger:
            raise InvalidCode('{} is not a valid trigger code for ({}) connector'.format(trigger_code, self.code))

        service_type = self.service.types.get(trigger_code)
        enhancer_configs = self.get_enhancers_config(service_type.datatype.code)
        process_configs = self.get_processes_config(service_type.datatype.code)

        # TODO: Save and load last trigger time, json file in project
        # TODO: Lock trigger if trigger is already running
        for object_data in service_type.trigger(TriggerRequest({}, datetime.datetime.now())):
            enhancers = []
            for enhancer_config in enhancer_configs:
                enhancers.append(Enhancer(
                    connector=enhancer_config['connector'],
                    mapping=enhancer_config['mapping'],
                    attributes=enhancer_config['attributes']
                ))

            data = DataMapObject(
                datatype=service_type.datatype,
                mapping=self.mapping.get(service_type.datatype.code, {}),
                enhancers=enhancers,
                data=object_data
            )

            # TODO: Make queue per processes and use multiple threads to call process function.
            # TODO: dont call process from own connector trigger

            for process_config in process_configs:
                logger.info('Processing {}:{} for object {}'.format(
                    process_config['connector'].code,
                    service_type.datatype.code,
                    data
                ))
                try:
                    object_id = process_config['connector'].service.types.get(service_type.datatype.code).process(
                        ProcessRequest(
                            connector=process_config['connector'],
                            config={},
                            data_object=data
                        ))
                    logger.info(' - Object id for ({}) is {}'.format(data, object_id))
                except Exception as e:
                    logger.error('Error processing data for process ({}:{}): {}'.format(
                        process_config['connector'].code,
                        service_type.datatype.code,
                        str(e)
                    ))

    def get(self, type_code, object_id):
        """
        Get data from service for give type and id

        :param str type_code:
        :param str object_id:
        :return:
        """
        service_type = self.service.types.get(type_code)
        if not service_type:
            raise Exception('There is no service type for given type code')

        # TODO: add id and generic_id to mapping type config
        object_data = service_type.get(GetRequest(object_id, generic_id='stub'))

        enhancers = []
        for enhancer_config in self.get_enhancers_config(type_code):
            enhancers.append(Enhancer(
                connector=enhancer_config['connector'],
                mapping=enhancer_config['mapping'],
                attributes=enhancer_config['attributes']
            ))

        return DataMapObject(
            datatype=service_type.datatype,
            mapping=self.mapping.get(service_type.datatype.code, {}),
            enhancers=[],  # todo add enhancers
            data=object_data
        )

    def get_enhancers_config(self, type_code):
        """
        Get list of enhancer configs

        :param str type_code:
        :return list:
        """
        # TODO: cache configs per type, maybe pre load on application
        configs = []
        for connector in self.app.connectors.values():
            mapping = connector.mapping.get(type_code, {})
            for enhancer_config in connector.enhancers:
                if enhancer_config.get('type') == type_code:
                    configs.append({
                        'connector': connector,
                        'mapping': mapping,
                        'attributes': enhancer_config.get('attributes', {})
                    })
        return configs

    def get_processes_config(self, type_code):
        """
        Get a list of process configs

        :param str type_code:
        :return list:
        """
        configs = []
        for connector in self.app.connectors.values():
            for process_config in connector.processes:
                if process_config.get('type') == type_code:
                    configs.append({
                        'connector': connector,
                        'attributes': process_config.get('attributes')
                    })
        return configs


class ConnectorConfig:
    """Connector config parser class"""

    def parse(self, config):
        """
        Parse given config and return dict with values for Connector __init__

        :param dict config:
        :return dict:
        """
        self.validate(config)
        service = ServiceFactory.get(config.get('service'))

        return {
            'code': config.get('code'),
            'service': service,
            'verbose_name': config.get('verbose_name'),
            'service_config': config.get('service_config'),
            'mapping': self.parse_mapping(service, config),
            'triggers': self.parse_triggers(service, config),
            'enhancers': config.get('enhancers'),
            'processes': config.get('processes')
        }

    def validate(self, config):
        """
        Validate base config

        :param str config:
        """
        if not config.get('code'):
            raise InvalidConnectorConfig('Config must have a code!')

        if not ServiceFactory.get(config.get('service')):
            raise InvalidConnectorConfig('Config is missing or has invalid service code')

        self.validate_mapping(config.get('mapping'))

    def validate_mapping(self, mapping):
        """
        Validate connector mapping

        :param dict mapping:
        :raises InvalidConnectorConfig:
        """
        valid = True

        if type(mapping) is not dict:
            raise InvalidConnectorConfig('Mapping config is not an dict')

        for type_code, config in mapping.items():
            pass

        return valid

    def parse_mapping(self, service, config):
        """
        Parse mapping

        :param xenops.service.Service service:
        :param dict config:
        :return dict:
        """
        mappings = {}

        # set default mappings
        for type_code, type_service in service.types.items():
            type_mapping = {}
            for converter in type_service.mapping:
                # @TODO Validate if attribute code exists in type
                if self.valid_converter_class(converter):
                    type_mapping[converter.attribute] = converter
                else:
                    logger.warning(
                        'Mapping value for ({}) on ({}) is not a valid converter class'.format(converter, type_code))
            mappings[type_code] = type_mapping

        # loop conn mapping
        for type_code, mapping in config.get('mapping', {}).items():
            service_type = service.types.get(type_code)

            if not service_type:
                logger.warning('could not create mapping for ({}), type does not exits for service'.format(type_code))
                continue

            type_mapping = mappings.get(type_code, {})
            if mapping.get('type', 'merge') != 'merge':
                type_mapping = {}

            for converter in mapping.get('attributes', []):
                # @TODO Validate if attribute code exists in type
                if self.valid_converter_class(converter):
                    type_mapping[converter.attribute] = converter
                else:
                    logger.warning(
                        'Mapping value for ({}) on ({}) is not a valid converter class'.format(converter, type_code))

        return mappings

    def valid_converter_class(self, value):
        """
        Validate if given object is an converter object

        :param str value:
        :return bool:
        """
        try:
            return isinstance(value, BaseConverter)
        except:  # noqa E722
            return False

    def parse_triggers(self, service, config):
        """
        Parse trigger config

        :param xenops.service.Service service:
        :param dict config:
        :return dict:
        """
        triggers = {}

        for trigger in config.get('triggers', []):
            if type(trigger) is dict:
                if service.types.get(trigger.get('type')):
                    triggers[trigger.get('type')] = trigger  # @TODO Add trigger config validation.
                else:
                    logger.warning(
                        "Could not create trigger for type ({}),"
                        " because trigger type does not exists for ({}) service".format(
                            trigger.get('type'),
                            service.code,
                        ))
            else:
                logger.error('Invalid trigger config supplied, only string or dictionary are valid!')

        return triggers



"""
xenops.service
~~~~~~~~~~~~~~

:copyright: 2017 by Maikel Martens
:license: GPLv3
"""
import logging

from xenops.data import DataTypeFactory

logger = logging.getLogger(__name__)


class TriggerRequest:
    """Trigger request"""

    def __init__(self, config, last_run):
        """
        Init trigger request

        :param dict config:
        :param datetime last_run:
        """
        self.config = config
        self.last_run = last_run


class GetRequest:
    """Get request"""

    def __init__(self, config, object_id, generic_id):
        """
        Init get request

        :param dict config:
        :param str object_id:
        :param str generic_id:
        """
        self.config = config
        self.object_id = object_id
        self.generic_id = generic_id


class ProcessRequest:
    """Process request"""

    def __init__(self, connector, config, data_object):
        """
        Init process request

        :param xenops.connector.Connector connector:
        :param dict config:
        :param xenops.data.DataMapObject data_object:
        """
        self.connector = connector
        self.config = config
        self.data_object = data_object


class Service:
    """Service class"""

    def __init__(self, code, verbose_name, service_types):
        """
        Init Service

        :param str code:
        :param str verbose_name:
        :param dict service_types:
        """
        self.code = code
        self.verbose_name = verbose_name
        self.types = service_types

    # TODO: Maybe add trigger/get/process(type_code, request)


class ServiceType:
    """Service type"""

    def __init__(self, datatype, mapping, trigger, get, process):
        """
        Init Service type

        :param xenops.data.DataType datatype:
        :param dict mapping:
        :param Callable trigger:
        :param Callable get:
        :param Callable process:
        """
        self.datatype = datatype
        self.mapping = mapping  # TODO: Move mapping to config, mapping is not used here
        self.trigger_function = trigger
        self.get_function = get
        self.process_function = process

    def trigger(self, request):
        """
        Run trigger

        :param xenops.service.TriggerRequest request:
        :return Generator[dict]:
        """
        for data in self.trigger_function(request):
            yield data

    def get(self, request):
        """
        Get data from service

        :param xenops.service.GetRequest request:
        :return dict:
        """
        return self.get_function(request)

    def process(self, request):
        """
        Process data from trigger

        :param xenops.service.ProcessRequest request:
        :return int: data object id from service
        """
        return self.process_function(request)


class ServiceFactory:
    """Service factory for registering and getting services"""

    _services = {}

    @classmethod
    def register(cls, config):
        """
        Register a service

        :param dict config:
        :return bool:
        """
        if not config.get('code'):
            logger.error('Service config must have a code!')
            return False

        if not config.get('type'):
            logger.error('Service must have at least one type!')

        types = {}
        for type_code, type_config in config.get('type').items():
            datatype = DataTypeFactory.get(type_code)

            if not datatype:
                logger.error('Service ({}) want to register data type ({}) that does not exists!'.format(config['code'],
                                                                                                         type_code))
                continue

            def trigger_function(*args, **kwargs):
                pass

            def get_function(*args, **kwargs):
                pass

            def process_function(*args, **kwargs):
                pass

            if 'trigger' in type_config and callable(type_config['trigger']):
                trigger_function = type_config['trigger']  # noqa F811

            if 'get' in type_config and callable(type_config['get']):
                get_function = type_config['get']  # noqa F811

            if 'process' in type_config and callable(type_config['process']):
                process_function = type_config['process']  # noqa F811

            types[datatype.code] = ServiceType(
                datatype=datatype,
                mapping=type_config.get('mapping', {}),  # @TODO validate mapping
                trigger=trigger_function,
                get=get_function,
                process=process_function
            )

        cls._services[config['code']] = Service(
            code=config['code'],
            verbose_name=config.get('verbose_name'),
            service_types=types
        )

        return True

    @classmethod
    def get(cls, code):
        """
        Get a service based on a code

        :param str code:
        :return xenops.service.Service:
        """
        return cls._services.get(code)



"""
xenops.cli
~~~~~~~~~~

:copyright: 2017 by Maikel Martens
:license: GPLv3
"""
import sys
import logging

from xenops.app import Application

logger = logging.getLogger()


class Command:
    """Xenops commandline"""

    def __init__(self, argv):
        """Init Setup Xenops"""
        logging.basicConfig(format="%(levelname)-8s: %(message)s", level=logging.DEBUG)

        # TODO add argparse

        # add option show flows: will show trigger -> enhances -> processes

        # Start app
        app = Application()
        app.trigger('product', 'pim_live')


if __name__ == '__main__':
    Command(sys.argv)



from setuptools import setup, find_packages


def readme():
    with open('README.rst') as f:
        return f.read()


setup(
    name='Xenops',
    version="0.0.1",
    description="Xenops is a simple program to sync data like (customers/products) between different systems",
    long_description=readme(),
    packages=find_packages(),
    include_package_data=True,
    classifiers = [
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
    ],
    author="Maikel Martens",
    author_email = 'maikel@martens.me',
    license='GPL3',
    url = 'https://github.com/krukas/Xenops',
    download_url = 'https://github.com/krukas/Xenops/releases/tag/0.0.1',
    keywords = ['Xenops', 'Connector', 'Enterprise Service Bus', 'ESB'],
    install_requires = [],
)
